/* cfs.c -- core routines for Coda inode manipulation
   Copyright 1997-98 Michael Callahan
   Copyright 1999 Carnegie Mellon University
   This program is free software.  You may copy it according
   to the conditions of the GNU General Public License version 2.0
   or later; see the file COPYING in the source directory.
*/


#include "MC.h"
#include <string.h>
#pragma intrinsic(strcpy, strlen, memcpy)

#pragma VxD_LOCKED_DATA_SEG
#pragma VxD_LOCKED_CODE_SEG

struct cnode *cnode_cache[CNODE_CACHESIZE];

int pathelement_lookup (PathElement *pe, struct cnode *parent, struct cnode **dir)
{
	USHORT p[MAX_PATH +1];
	int res = 0;
	ViceFid fid;
	int vtype;
	int i;

	for (i = 0; i < (pe->pe_length /2 -1); i++)
		p[i] = pe->pe_unichars[i];
	p[i] = '\0';
	res = venus_lookup(parent->c_fid, (unsigned short *)&p, &fid, &vtype, CLU_CASE_INSENSITIVE, pe->pe_length / 2, 0);
	
	if (!res) {

		*dir = iget(fid);
		if (!dir) 
			res = ERROR_NOT_ENOUGH_MEMORY;
		else 
			(*dir)->c_vattr.va_type = vtype;
	}

error:
	if (res) 
		printf ("pathelement_lookup: failure %x", res);
	return res;
}

int open_namei (ParsedPath *pp, struct cnode **pcnode)
{
	struct cnode *parent, *dir;
	PathElement *pe;
	int res = 0;

	if (!root) {
		logmsg ("ERROR: open_namei: no root cnode!");
		return ERROR_DEV_NOT_EXIST;
	}

	if (IFSIsRoot(pp)) {
		logmsg ("ERROR: open_namei called on '\\' path");
		return ERROR_DEV_NOT_EXIST;
	}

	dir = iget(root->c_fid);
	pe = pp->pp_elements;
	while (!res && (pe != IFSLastElement(pp)) ) {
		parent = dir;
		res = pathelement_lookup (pe, parent, &dir);
		iput (parent);
		if (!res && dir->c_vattr.va_type != C_VDIR) {
			iput (dir);
			res = ERROR_PATH_NOT_FOUND;
			logmsg ("found non-dir path entry");
		}
		pe = IFSNextElement(pe);
	}

	if (!res)
		(*pcnode) = dir;
	if (res == ERROR_FILE_NOT_FOUND)
		res = ERROR_PATH_NOT_FOUND;
	return res;
}

int namei (ParsedPath *pp, USHORT *uFName, struct cnode **pcnode)
{
	struct cnode *dircnode = NULL;
	int res = 0;
	ViceFid fid;
	DWORD type;

	if (IFSIsRoot(pp)) {
		fid = root->c_fid;
		type = root->c_vattr.va_type;
	} else {
		res = open_namei(pp, &dircnode);
		if (res)
			return res;

		logucode(uFName);
		res = lookup_ucode (dircnode->c_fid, uFName, &fid, &type);
		iput (dircnode);
		if (res)
			return res;
	}

	*pcnode = iget (fid);
	if (!*pcnode)
		return ERROR_NOT_ENOUGH_MEMORY;

	(*pcnode)->c_vattr.va_type = type;
	return 0;
}

int ascii_namei (PCHAR p, struct cnode **pcnode)
{
	PCHAR end;
	ViceFid fid;
	int res = 0, lasttime;
	DWORD dwType;

	fid = root->c_fid;
	if (p) lasttime = 0; else lasttime = 1;
	logmsg(p);
	if (*p == '.') p++;
	if (*p == '\\' || *p == '/') p++;
	while (!lasttime && *p != '\0') {
		for (end = p;*end && *end != '/' && *end !='\\'; end++)
			;
		if (!*end)
			lasttime++;
		else
			*end = '\0';

		logmsg (p);
		res = lookup (fid, p, &fid, &dwType, CLU_CASE_INSENSITIVE);
		if (res)
			break;
		p = end + 1;
	}
	if (!res)
		*pcnode = iget(fid);
	return res;
}


struct cnode *iget (ViceFid fid)
{
	struct cnode *cn = NULL;
	struct cnode *newcn = NULL;

	/* this loop executes at most twice -- first time to look for an
	   existing cnode, second time to check one hasn't shown up
	   while we were allocating the memory */
	while (1) {
		_asm pushf
		_asm cli
		for (cn = cnode_cache[cnode_hash(&fid)];
			 cn; 
			 cn = cn->c_next) {
			if (cn->c_fid.Volume == fid.Volume &&
				cn->c_fid.Vnode == fid.Vnode &&
				cn->c_fid.Unique == fid.Unique)
				break;
		}

		if (cn) {
			cn->c_ocount++;
			_asm popf
			break;
		} else if (newcn) {
			newcn->c_next = cnode_cache[cnode_hash(&fid)];
			cnode_cache[cnode_hash(&fid)] = newcn;
			_asm popf
			/* indicate that we got the cnode, and zero
			   newcn so that we don't free it */
			cn = newcn;
			newcn = NULL;
			break;
		}
	
		_asm popf

		newcn = _HeapAllocate(sizeof(struct cnode), HEAPZEROINIT);
		if (!newcn)
			return NULL;
		newcn->c_ocount++;
		newcn->c_fid = fid;
		newcn->c_magic = CODA_CNODE_MAGIC;
	}
	if (newcn) {
		/* turns out we allocated it, didn't need it */
		_HeapFree(newcn, 0);
	}
	return cn;
}


void iput (struct cnode *cnode)
{
	struct cnode **pcn;
	cnode->c_ocount--;
	if (cnode->c_ocount == 0) {
		/* remove from list atomically */
		_asm pushf
		_asm cli
		if (cnode->c_ocount == 0) {
			for (pcn = &cnode_cache[cnode_hash(&cnode->c_fid)];
				 *pcn;
				 pcn = &((*pcn)->c_next)) {
					if (*pcn == cnode) {
						*pcn = cnode->c_next;
						break;
					}
			}
			_asm popf
			_HeapFree(cnode, 0);
			return;
		}
		_asm popf;
	}
}

int openVFid (ViceFid VFid, int flags, WORD access, WORD attr, BYTE action, 
			  struct request *inreq, PDWORD phdl, PDWORD ptakenaction,
			  PWORD pattr)
{
	char path[MAX_PATH];
	DWORD takenaction;
	int res = 0;

	logmsg("--openVFid--");

	*phdl = 0;
	res = venus_open_by_path(VFid, (char *)&path, inreq, flags);
	if (res) goto error;

	if (!ptakenaction)
		ptakenaction = &takenaction;

	res = R0_OpenCreateFile ((BOOL) TRUE,
		(WORD) (access & ACCESS_MODE_MASK),
		attr,
		ACTION_EXISTS_OPEN/*action*/,
		(BYTE) 0,
		path,
		phdl,
		ptakenaction);

	if (res) {
		printf ("R0_OpenCreate failed in openVFid %x", res);
		goto error;
	}

	/* sometimes we need attributes of opened file */
	if (pattr) {
		res = R0_GetFileAttributes(path, pattr);
		if (res) {
			logmsg ("GetFileAttributes failed");
			goto error;
		}
	}
	printf("open handle: %x", *phdl);

error:
	if (res && *phdl)
		R0_CloseFile(*phdl);
	logmsg("--end of openVFid--");
	return res;
}

int lookup_ucode (ViceFid dirfid, USHORT *name, ViceFid *fid, DWORD *type)
{
	int len, res;

	logmsg("--lookup_ucode--");

	for (len = 0; name[len]; len++)
		;
	len++;

	res = venus_lookup(dirfid, name, fid, type, CLU_CASE_INSENSITIVE, len, 0);

	printf ("<%x %x %x> type %x", fid->Volume, fid->Vnode, fid->Unique,
		*type);

	logmsg("--end of lookup_ucode--");
	return res;
}


int lookup (ViceFid dirfid, PCHAR name, ViceFid *fid, DWORD *type, int flags)
{
	int res, len;
	USHORT uniname[MAX_PATH];

	logmsg("--lookup--");

	len = strlen(name)+1;
	BCSToUni(uniname, name, len, BCS_OEM);
	res = venus_lookup(dirfid, uniname, fid, type, flags, len, 0);

	printf ("<%x %x %x> type %x", fid->Volume, fid->Vnode, fid->Unique,
		*type);

	logmsg("--end of lookup--");
	return res;
}

int getattr (ViceFid fid, struct coda_vattr *vattr)
{
	int res;

	logmsg("--getattr--");
	res = venus_getattr(fid, vattr);
	logmsg("--end of getattr--");

	return res;
}

int coda_getattr (struct cnode *cn)
{
	int res = 0;
	if (VALID_VATTR(cn))
		return 0;
	res = getattr (cn->c_fid, &cn->c_vattr);
	if (res == 0)
		cn->c_flags |= C_VATTR;
	return res;
}

int setattr(ViceFid fid, struct coda_vattr vattr){
	int res;
	
	logmsg("--setattr--");
	res = venus_setattr(fid, vattr);
	logmsg("--end of setattr--");

	return res;

}
int coda_setattr (struct cnode *cn)
{
	int res = 0;

	res = setattr(cn->c_fid, cn->c_vattr);

	return res;
}
