/* file.c -- Windows 95 IFS routines for manipulating open files
   Copyright 1997-98 Michael Callahan
   Copyright 1999 Carnegie Mellon University
   This program is free software.  You may copy it according
   to the conditions of the GNU General Public License version 2.0
   or later; see the file COPYING in the source directory.
*/

#include <string.h>
#pragma intrinsic (strcpy,strlen, memcpy)

#include "mc.h"
#include <vmm.h>
#include "vmmextra.h"

#pragma VxD_LOCKED_DATA_SEG
#pragma VxD_LOCKED_CODE_SEG

static struct hndlmisc hf_tbl = {
	IFS_VERSION,
	IFS_REVISION,
	NUM_HNDLMISC,

	FileSeek,
	CloseFile,
	CommitFile,
	LockFile,				/* LockFile,*/
	FileDateTime,			/* FileDateTime, */
	NamedPipeRequest,		/* NamedPipeRequest */
	NamedPipeHandleInfo,	/* NamedPipeHandleInfo, */
	EnumerateHandle			/* EnumerateHandle */
};


#define filesig 0xdeaffade
struct fileinfo {
	DWORD	signature;
	DWORD	memctx;
	DWORD	hdl;
	DWORD	size;
	DWORD	flags;
	ParsedPath *ppath;
	struct cnode *cn;
};

int OpenFile (pioreq pir)
{
	DWORD takenaction, hdl = 0;
	WORD attr;
	int res = 0;
	int flags, dummy;
	struct cnode *dircnode = NULL;
	ViceFid fid;
	struct fileinfo *pfi = NULL;

	logmsg ("--open--");
	logucode (pir->ir_upath);
	logucode (pir->ir_uFName);

	pir->ir_fh = NULL;
	hdl = 0;

	res = open_namei(pir->ir_ppath, &dircnode);
	if (res) {
		printf ("open: open_namei failed %x", res);
		goto error;
	}

	/* first, either look up inode or create new inode */

	/* first, if the request flags allow it, we try to open existing file */
	if (pir->ir_options & ACTION_EXISTS_OPEN) {

		logmsg ("--open : ACTION_EXISTS_OPEN");

		res = venus_lookup(dircnode->c_fid, pir->ir_uFName, &fid, &dummy, CLU_CASE_INSENSITIVE, MAX_PATH+1, 1);

		if (!res) goto gotit;
	}

	/* if we get here, then check if flags demand that we create a new file */
	if (pir->ir_options & ACTION_NEXISTS_CREATE){
		logmsg ("--open : ACTION_NEXISTS_CREATE");
		printf ("--open flags %x", pir->ir_options);
		res = venus_create(dircnode, pir, &fid);
	}
	
	if (res) goto error;	

gotit:

	flags = 0;
	switch (pir->ir_flags & ACCESS_MODE_MASK) {
	case ACCESS_READONLY:
	case ACCESS_EXECUTE:
		flags = C_O_READ;
		break;

	case ACCESS_WRITEONLY:
		flags = C_O_WRITE;
		break;

	case ACCESS_READWRITE:
		flags = C_O_READ | C_O_WRITE;
		break;

	default:
		printf ("Unknown access mode %x", pir->ir_flags & ACCESS_MODE_MASK);
	}
	if (pir->ir_options & ACTION_TRUNCATE)
		flags |= C_O_TRUNC;
	
	printf("about to enter openVFid: Access: %x, Action: %x", (DWORD)(pir->ir_flags & ACCESS_MODE_MASK), (DWORD)(pir->ir_options & ACTION_MASK));
	res = openVFid(fid, flags, (WORD) (pir->ir_flags & ACCESS_MODE_MASK),
				(WORD) pir->ir_attr, (BYTE) (pir->ir_options & ACTION_MASK),
				/*req*/NULL, &hdl, &takenaction, &attr);
	if (res) {
		logmsg ("open: OpenCreate failed");
		goto error;
	}

	printf ("takenaction %x attr %x", takenaction, attr);

	pir->ir_options = (WORD) takenaction;
	pir->ir_attr = attr;

	// ir_size should get file size
	pir->ir_size = 0;
	if (pir->ir_options == ACTION_OPENED) {
		printf ("getfilesize (hdl %x)", hdl);
		res = R0_GetFileSize(hdl, &pir->ir_size);
		if (res) {
			logmsg ("GetFileSize failed");
			goto error;
		}
	}

	SetHandleFunc(pir, ReadFile, WriteFile, &hf_tbl);

	pfi = (struct fileinfo *) _HeapAllocate (sizeof(struct fileinfo), 0);
	if (!pfi) {
		res = ERROR_NOT_ENOUGH_MEMORY;
		goto error;
	}

	pfi->signature = filesig;
	pfi->memctx = (DWORD) _GetCurrentContext();
	pfi->hdl = hdl;
	pfi->flags = flags;
	pfi->cn = iget(fid);
	if (!pfi->cn) {
		res = ERROR_NOT_ENOUGH_MEMORY;
		goto error;
	}
	pfi->cn->c_vattr.va_type = C_VREG;
	pfi->size = pir->ir_size;
	pfi->ppath = pir->ir_ppath;

	pir->ir_fh = pfi;
	
error:
	if (dircnode)
		iput(dircnode);
	if (res && hdl)
		R0_CloseFile(hdl);
	if (res && pfi)
		_HeapFree(pfi, 0);
	if (res)
		printf ("open failed: %x", res);
	else
		printf ("end open (%x) ctx %x", (int) pfi, _GetCurrentContext());

	logmsg ("--end of open--");
	return pir->ir_error = res;
}

void checkctx (struct fileinfo *pfi)
{
	if (pfi->memctx != (DWORD) _GetCurrentContext()) {
		printf ("CONTEXT ERROR: orig %x now %x",
			pfi->memctx, (DWORD) _GetCurrentContext());
	}
}

int ReadFile (pioreq pir)
{
	struct fileinfo *pfi = (struct fileinfo *) pir->ir_fh;
	int res;

	printf ("--read-- (%x) pos %x len %x addr %x", (int) pfi, pir->ir_pos, pir->ir_length, pir->ir_data);

	if (pfi->signature != filesig) {
		logmsg ("BAD SIGNATURE!");
		return pir->ir_error = ERROR_INVALID_PARAMETER;
	}

	checkctx (pfi);

	if (pir->ir_pos >= pfi->size)
		pir->ir_length = 0;
	else if (pir->ir_pos + pir->ir_length > pfi->size)
		pir->ir_length = pfi->size - pir->ir_pos;

	res = R0_ReadFile(TRUE,
			pfi->hdl,
			pir->ir_length,
			pir->ir_pos,
			(PBYTE) pir->ir_data,
			(PDWORD) &pir->ir_length);
	if (!res) {
		pir->ir_pos += pir->ir_length;
	}
	pir->ir_error = res;
	printf ("end read (err %x)", res);
	logmsg("--end of read--");
	return pir->ir_error = res;
}

int WriteFile (pioreq pir)
{
	struct fileinfo *pfi = (struct fileinfo *) pir->ir_fh;
	int res;

	printf ("--write-- (%x) pos %x len %x", (DWORD) pfi, pir->ir_pos, pir->ir_length);

	if (pfi->signature != filesig) {
		logmsg ("BAD SIGNATURE!");
		return pir->ir_error = ERROR_INVALID_PARAMETER;
	}

	checkctx (pfi);

	res = R0_WriteFile(TRUE,
			pfi->hdl,
			pir->ir_length,
			pir->ir_pos,
			(PBYTE) pir->ir_data,
			(PDWORD) &pir->ir_length);
	if (!res) {
		pir->ir_pos += pir->ir_length;
		if (pir->ir_pos > pfi->size)
			pfi->size = pir->ir_pos;
	}

	pir->ir_error = res;
	printf ("end write (err %x)", res);
	logmsg("--end of write--");
	return pir->ir_error = res;
}

int FileSeek (pioreq pir)
{
	struct fileinfo *pfi = (struct fileinfo *) pir->ir_fh;
	int res = 0;

	printf ("--seek-- (%x) flags %x pos %x", 
		(int) pfi, pir->ir_flags, pir->ir_pos);

	if (pfi->signature != filesig) {
		logmsg ("BAD SIGNATURE!");
		return pir->ir_error = ERROR_INVALID_PARAMETER;
	}
	checkctx(pfi);

	if (pir->ir_flags == FILE_END) {
		DWORD dwSize;
		printf ("seek: about to do GetFileSize (hdl %x)", pfi->hdl);
		res = R0_GetFileSize (pfi->hdl,	&dwSize);
		if (!res) {
			pir->ir_pos += dwSize;
			printf ("seek: pos now %x", pir->ir_pos);
		}
	}
	
	printf ("--end of seek (err %x)--", res);
	return pir->ir_error = res;
}

int CloseFile (pioreq pir)
{
	struct request *req = NULL;
	struct fileinfo *pfi = (struct fileinfo *) pir->ir_fh;
	int res = 0;

	printf ("--close-- (%x) closetype %x", (int) pfi, pir->ir_flags);
	if (pfi->signature != filesig) {
		logmsg ("BAD SIGNATURE!");
		return pir->ir_error = ERROR_INVALID_PARAMETER;
	}
	checkctx (pfi);

	if (pir->ir_flags != CLOSE_FINAL) {
		printf ("end close (err %x)", res);
		return pir->ir_error = res;
	}
	
	res = R0_CloseFile(pfi->hdl);
	/* XXX: what to do if there is an error here? */
	if (res !=0 ){
		printf ("Error at R0_CloseFile in CloseFile");
		logmsg(" -- close error ");
	}

	res = venus_close(pfi->cn->c_fid, pfi->flags);

error:
	if (req)
		FreeRequest (req);
	iput (pfi->cn);
	_HeapFree (pfi, 0);

	printf ("--end of close (err %x)--", res);
	return pir->ir_error = res;
}

int CommitFile (pioreq pir)
{
	printf ("--commit-- (%x)", pir->ir_fh);
	logmsg("--end of commit--");
	return pir->ir_error = 0;
}

int LockFile (pioreq pir)
{
	struct fileinfo *pfi = (struct fileinfo *) pir->ir_fh;
	printf ("--lock-- (%x) flags %x opt %x", (DWORD) pfi, pir->ir_flags, pir->ir_options);
	if (pfi->signature != filesig) {
		logmsg ("BAD SIGNATURE!");
		return pir->ir_error = ERROR_INVALID_PARAMETER;
	}
	checkctx (pfi);
	logmsg("--end of lock--");
	return pir->ir_error = 0;
}

int FileDateTime (pioreq pir)
{
	struct fileinfo *pfi = (struct fileinfo *) pir->ir_fh;
	int res = ERROR_INVALID_PARAMETER;

	printf ("--filedatetime-- (%x) subfunc %x", (DWORD) pfi, pir->ir_flags);
	if (pfi->signature != filesig) {
		logmsg ("BAD SIGNATURE!");
		return pir->ir_error = ERROR_INVALID_PARAMETER;
	}
	checkctx (pfi);

	res = coda_getattr(pfi->cn);
	if (res)
		goto error;
	
	switch (pir->ir_flags) {
	case SET_MODIFY_DATETIME:
		pfi->cn->c_vattr.va_mtime.ts_sec = IFSMgr_DosToNetTime(pir->ir_dostime);
		pfi->cn->c_vattr.va_mtime.ts_nsec = 0;
		break;

	case GET_MODIFY_DATETIME:
		pir->ir_dostime = IFSMgr_NetToDosTime(pfi->cn->c_vattr.va_mtime.ts_sec);
		break;

	case SET_LAST_ACCESS_DATETIME:
		pfi->cn->c_vattr.va_atime.ts_sec = IFSMgr_DosToNetTime(pir->ir_dostime);
		pfi->cn->c_vattr.va_atime.ts_nsec = 0;
		break;

	case GET_LAST_ACCESS_DATETIME:
		pir->ir_dostime = IFSMgr_NetToDosTime(pfi->cn->c_vattr.va_atime.ts_sec);
		break;

	case SET_CREATION_DATETIME:
		pfi->cn->c_vattr.va_ctime.ts_sec = IFSMgr_DosToNetTime(pir->ir_dostime);
		pfi->cn->c_vattr.va_ctime.ts_nsec = pir->ir_options * 10000000;
		break;

	case GET_CREATION_DATETIME:
		pir->ir_dostime = IFSMgr_NetToDosTime(pfi->cn->c_vattr.va_ctime.ts_sec);
		pir->ir_options = pfi->cn->c_vattr.va_ctime.ts_nsec / 10000000;
		break;
	}

	if (pir->ir_flags == SET_MODIFY_DATETIME ||
		pir->ir_flags == SET_LAST_ACCESS_DATETIME ||
		pir->ir_flags == SET_CREATION_DATETIME)
		res = coda_setattr(pfi->cn);

error:
	printf ("--end of filedatetime (err %x)--", res);
	return pir->ir_error = res;
}

int NamedPipeRequest (pioreq pir)
{
	printf ("FAILING -- NamedPipeRequest %x", pir->ir_flags);
	return pir->ir_error = ERROR_INVALID_FUNCTION;
}

int NamedPipeHandleInfo (pioreq pir)
{
	logmsg ("FAILING -- NamedPipeHandleInfo");
	return pir->ir_error = ERROR_INVALID_FUNCTION;
}

int EnumerateHandle (pioreq pir)
{
	struct fileinfo *pfi = (struct fileinfo *) pir->ir_fh;
	int res = ERROR_INVALID_FUNCTION;

	logmsg ("--enumeratehandle--");

	checkctx (pfi);

	switch (pir->ir_flags) {
	case ENUMH_GETFILEINFO:
		logmsg ("--enumeratehandle: getfileinfo");
		res = coda_getattr(pfi->cn);
		if (res)
			break;			

		((_BY_HANDLE_FILE_INFORMATION *)pir->ir_data)->bhfi_dwFileAttributes = (pfi->cn->c_vattr.va_type == C_VDIR) ? FILE_ATTRIBUTE_DIRECTORY : 0;
		((_BY_HANDLE_FILE_INFORMATION *)pir->ir_data)->bhfi_ftCreationTime = IFSMgr_NetToWin32Time(pfi->cn->c_vattr.va_ctime.ts_sec);
		((_BY_HANDLE_FILE_INFORMATION *)pir->ir_data)->bhfi_ftLastAccessTime = IFSMgr_NetToWin32Time(pfi->cn->c_vattr.va_atime.ts_sec);
		((_BY_HANDLE_FILE_INFORMATION *)pir->ir_data)->bhfi_ftLastWriteTime = IFSMgr_NetToWin32Time(pfi->cn->c_vattr.va_mtime.ts_sec);
		((_BY_HANDLE_FILE_INFORMATION *)pir->ir_data)->bhfi_dwVolumeSerialNumber = 0;
		((_BY_HANDLE_FILE_INFORMATION *)pir->ir_data)->bhfi_nFileSizeHigh = 0;
		((_BY_HANDLE_FILE_INFORMATION *)pir->ir_data)->bhfi_nFileSizeLow = (unsigned long) pfi->cn->c_vattr.va_size;
		((_BY_HANDLE_FILE_INFORMATION *)pir->ir_data)->bhfi_nNumberOfLinks = (unsigned long) pfi->cn->c_vattr.va_nlink;
		((_BY_HANDLE_FILE_INFORMATION *)pir->ir_data)->bhfi_nFileIndexHigh = 0;
		((_BY_HANDLE_FILE_INFORMATION *)pir->ir_data)->bhfi_nFileIndexLow = 0;

		break;

	case ENUMH_GETFILENAME:
		logmsg ("--enumeratehandle: getfilename");
		if (pfi && pfi->ppath) 
			memcpy ((PBYTE) pir->ir_ppath, pfi->ppath, IFSPathSize(pfi->ppath));
		res = 0;
		break;

	case ENUMH_GETFINDINFO:
		logmsg ("--enumeratehandle: getfindinfo");
		if (pfi && pfi->ppath) 
			memcpy ((PBYTE) pir->ir_ppath, pfi->ppath, IFSPathSize(pfi->ppath));
		res = 0;
		break;

	case ENUMH_RESUMEFIND:
		logmsg ("     resumefind");
		break;

	case ENUMH_RESYNCFILEDIR:
		logmsg ("     resyncfiledir");
		break;

	default:
		logmsg ("     unknown subfunc code!");
		break;
	}

	logmsg ("--end of enumeratehandle--");
	return pir->ir_error = res;
}

